from .archive_exporter import ArchiveExporter, NothingToExport
from .column_handlers import ColumnApply, ColumnMap
from .constants import SYSTEM_COLUMNS
from .csv_exporters import CsvCrfTablesExporter, CsvCrfInlineTablesExporter
from .csv_exporters import CsvNonCrfTablesExporter
from .csv_exporters import CsvModelExporter, CsvExporter, CsvTablesExporter
from .database import Database
from .df_handlers import CrfDfHandler, NonCrfDfHandler, DfHandler
from .dialects import CrfDialect, MysqlDialect, RsDialect
from .model_to_dataframe import ModelToDataframe, SubjectModelToDataframe
from .model_to_dataframe import ValueGetter
from .table_to_dataframe import TableToDataframe
from .tables import Aliquot, Consent, Visit, Requisition
from .utils import datetime_to_date
from .utils import decrypt, DecryptError
from .utils import identity256, identity256_decrypt
from .utils import undash





class CrfDialect:

    def __init__(self, obj=None):
        self.obj = obj

    @property
    def select_visit_and_related(self):
        """Returns an SQL statement that joins visit, appt, and registered_subject.
        """
        sql = (
            'SELECT R.subject_identifier, R.screening_identifier, R.dob, '
            'R.gender, R.subject_type, R.sid, '
            'V.report_datetime as visit_datetime, A.appt_status, '
            'A.visit_code, A.timepoint, V.reason, '
            'A.appt_datetime, A.timepoint_datetime,  '
            'R.screening_age_in_years, R.registration_status, R.registration_datetime, '
            'R.randomization_datetime, V.survival_status, V.last_alive_date, '
            f'V.id as {self.obj.visit_column} '
            f'from {self.obj.appointment_tbl} as A '
            f'LEFT JOIN {self.obj.visit_tbl} as V on A.id=V.appointment_id '
            f'LEFT JOIN {self.obj.registered_subject_tbl} as R '
            'on A.subject_identifier=R.subject_identifier '
        )
        return sql, None

    @property
    def select_visit_and_related2(self):
        """Returns an SQL statement that joins visit, appt, and registered_subject.

        This is for older EDC versions that use this schema.
        """
        sql = (
            'SELECT R.subject_identifier, R.screening_identifier, R.dob, '
            'R.gender, R.subject_type, R.sid, '
            'V.report_datetime as visit_datetime, A.appt_status, V.study_status, '
            'VDEF.code as visit_code, VDEF.title as visit_title, VDEF.time_point, V.reason, '
            'A.appt_datetime, A.timepoint_datetime, A.best_appt_datetime, '
            'R.screening_age_in_years, R.registration_status, R.registration_datetime, '
            'R.randomization_datetime, V.survival_status, V.last_alive_date, '
            f'V.id as {self.obj.visit_column} '
            f'from {self.obj.appointment_tbl} as  A '
            f'LEFT JOIN {self.obj.visit_tbl} as V on A.id=V.appointment_id '
            f'LEFT JOIN {self.obj.visit_definition_tbl} as VDEF '
            'on A.visit_definition_id=VDEF.id '
            f'LEFT JOIN {self.obj.registered_subject_tbl} as R '
            'on A.registered_subject_id=R.id '
        )
        return sql, None

    @property
    def select_inline_visit_and_related(self):
        sql = (
            'SELECT R.subject_identifier, R.screening_identifier, R.dob, '
            'R.gender, R.subject_type, R.sid, '
            'V.report_datetime as visit_datetime, A.appt_status, V.study_status, '
            'VDEF.code as visit_code, VDEF.title as visit_title, VDEF.time_point, V.reason, '
            'A.appt_datetime, A.timepoint_datetime, A.best_appt_datetime, '
            'R.screening_age_in_years, R.registration_status, R.registration_datetime, '
            'R.randomization_datetime, V.survival_status, V.last_alive_date, '
            f'V.id as {self.obj.visit_column} '
            f'from {self.obj.appointment_tbl} as A '
            f'LEFT JOIN {self.obj.visit_tbl} as V on A.id=V.appointment_id '
            f'LEFT JOIN {self.obj.visit_definition_tbl} as VDEF '
            'on A.visit_definition_id=VDEF.id '
            f'LEFT JOIN {self.obj.registered_subject_tbl} as R '
            'on A.registered_subject_id=R.id '
        )
        return sql, None



from .helper import Helper
from .missing_subject_identifiers import missing_subject_identifiers
from .model_to_dataframe import ModelToDataframe
from .subject_model_to_dataframe import SubjectModelToDataframe
from .value_getter import ValueGetter, ValueGetterInvalidLookup, ValueGetterUnknownField



from django.urls import path
from django.views.generic.base import RedirectView

from .admin_site import edc_pdutils_admin

app_name = 'edc_pdutils'

urlpatterns = [
    path('admin/', edc_pdutils_admin.urls),
    path('', RedirectView.as_view(url='admin/'), name='home_url'),
]




class MysqlDialect:

    def __init__(self, dbname=None):
        self.dbname = dbname

    def show_databases(self):
        sql = 'SELECT SCHEMA_NAME AS `database` FROM INFORMATION_SCHEMA.SCHEMATA'
        return sql, None

    def show_tables(self, app_label=None):
        params = {'dbname': self.dbname, 'app_label': f'{app_label}%%'}
        select = ('SELECT table_name FROM information_schema.tables')
        where = [f'table_schema=%(dbname)s']
        if app_label:
            where.append(f'table_name LIKE %(app_label)s')
        sql = f'{select} WHERE {" AND ".join(where)}'
        return sql, params

    def show_tables_with_columns(self, app_label=None, column_names=None):
        column_names = '\',\''.join(column_names)
        params = {
            'dbname': self.dbname,
            'app_label': f'{app_label}%%',
            'column_names': column_names}
        sql = (
            'SELECT DISTINCT table_name FROM information_schema.columns '
            f'WHERE table_schema=%(dbname)s '
            f'AND table_name LIKE %(app_label)s '
            f'AND column_name IN (%(column_names)s)')
        return sql, params

    def show_tables_without_columns(self, app_label=None, column_names=None):
        column_names = '\',\''.join(column_names)
        params = {
            'dbname': self.dbname,
            'app_label': f'{app_label}%%',
            'column_names': column_names}
        sql = ('SELECT DISTINCT table_name FROM information_schema.tables as T '
               f'WHERE T.table_schema = %(dbname)s '
               f'AND T.table_type = \'BASE TABLE\' '
               f'AND T.table_name LIKE %(app_label)s '
               'AND NOT EXISTS ('
               'SELECT * FROM INFORMATION_SCHEMA.COLUMNS C '
               'WHERE C.table_schema = T.table_schema '
               'AND C.table_name = T.table_name '
               f'AND C.column_name IN (%(column_names)s))')
        return sql, params

    def select_table(self, table_name=None):
        params = {'table_name': table_name}
        sql = f'select * from {table_name}'
        return sql, params

    def show_inline_tables(self, referenced_table_name=None):
        params = {'referenced_table_name': referenced_table_name}
        sql = (f'SELECT DISTINCT referenced_table_name, table_name, '
               'column_name, referenced_column_name '
               f'FROM information_schema.key_column_usage '
               f'where referenced_table_name=%(referenced_table_name)s')
        return sql, params



import pandas as pd

from ..dialects import CrfDialect
from ..constants import SYSTEM_COLUMNS
from .df_handler import DfHandler


class CrfDfHandlerError(Exception):
    pass


class CrfDfHandler(DfHandler):

    crf_dialect_cls = CrfDialect
    visit_tbl = None
    visit_column = None  # e.g.  'subject_visit_id'
    appointment_tbl = 'edc_appointment_appointment'
    registered_subject_tbl = 'edc_registration_registeredsubject'
    dialect_select_visit_and_related = 'select_visit_and_related'

    system_columns = SYSTEM_COLUMNS
    sort_by = ['subject_identifier', 'visit_datetime']
    exclude_export_columns = True
    exclude_system_columns = False
    exclude_columns = ['form_as_json']

    def __init__(self, exclude_system_columns=None, **kwargs):
        self._df_visit_and_related = pd.DataFrame()
        self.crf_dialect = self.crf_dialect_cls(self)
        self.exclude_system_columns = exclude_system_columns or self.exclude_system_columns
        super().__init__(**kwargs)

    def prepare_dataframe(self, **kwargs):
        """Merges CRF dataframe with df_visit_and_related
        on visit_column.
        """
        if not self.visit_column:
            raise CrfDfHandlerError(
                f'visit column cannot be None. '
                f'See class attr \'visit_column\' on {repr(self)}')
        self.dataframe = pd.merge(
            left=self.dataframe, right=self.df_visit_and_related,
            how='left', on=self.visit_column,
            suffixes=['_notused', ''])
        self.dataframe = self.dataframe[self.columns]

    @property
    def columns(self):
        """Returns a list of column names.
        """
        crf_columns = list(self.dataframe.columns)
        crf_columns.pop(crf_columns.index(self.visit_column))
        columns = list(self.df_visit_and_related.columns)
        columns.extend([c for c in crf_columns if c not in columns])
        columns = [col for col in columns if col not in self.exclude_columns]
        # "export_" columns
        if self.exclude_export_columns:
            columns = [col for col in columns if not col.startswith('export_')]

        if self.exclude_system_columns:
            columns = [col for col in columns if col not in self.system_columns]
        else:
            # "system" columns, move to the end
            columns = [col for col in columns if col not in self.system_columns]
            columns.extend(self.system_columns)
        return columns

    @property
    def df_visit_and_related(self):
        """Returns a dataframe of the crf_dialect's `select_visit_and_related`
        SQL statement.
        """
        if self._df_visit_and_related.empty:
            sql, params = getattr(
                self.crf_dialect, self.dialect_select_visit_and_related)
            self._df_visit_and_related = self.db.read_sql(sql, params=params)
        return self._df_visit_and_related



import numpy as np
import pandas as pd
import sys

from django.apps import apps as django_apps
from django.db.models.constants import LOOKUP_SEP

from .value_getter import ValueGetter


class ModelToDataframe:
    """
        m = ModelToDataframe(model='edc_pdutils.crf', add_columns_for='clinic_visit')
        my_df = m.dataframe
    """

    value_getter_cls = ValueGetter
    sys_field_names = ['_state', '_user_container_instance', 'using']
    edc_sys_columns = [
        'created', 'modified',
        'user_created', 'user_modified',
        'hostname_created', 'hostname_modified',
        'device_created', 'device_modified',
        'revision']

    def __init__(self, model=None, queryset=None, query_filter=None,
                 add_columns_for=None, decrypt=None, drop_sys_columns=None,
                 **kwargs):
        self._columns = None
        self._encrypted_columns = None
        self._dataframe = pd.DataFrame()
        self.drop_sys_columns = drop_sys_columns
        self.decrypt = decrypt
        self.add_columns_for = add_columns_for
        self.query_filter = query_filter or {}
        if queryset:
            self.model = queryset.model._meta.label_lower
        else:
            self.model = model
        self.queryset = queryset or self.model_cls.objects.all()

    @property
    def dataframe(self):
        """Returns a pandas dataframe.
        """
        if self._dataframe.empty:
            row_count = self.queryset.count()
            if row_count > 0:
                if self.decrypt and self.has_encrypted_fields:
                    sys.stdout.write(
                        f'   PII will be decrypted! ... \n')
                    queryset = self.queryset.filter(**self.query_filter)
                    data = []
                    for index, model_obj in enumerate(queryset.order_by('id')):
                        sys.stdout.write(
                            f'   {self.model} {index + 1}/{row_count} ... \r')
                        row = []
                        for lookup, column_name in self.columns.items():
                            value = self.get_column_value(
                                model_obj=model_obj, column_name=column_name, lookup=lookup)
                            row.append(value)
                        data.append(row)
                        self._dataframe = pd.DataFrame(
                            data, columns=self.columns)
                else:
                    columns = [
                        col for col in self.columns if col not in self.encrypted_columns]
                    queryset = self.queryset.values_list(
                        *columns).filter(**self.query_filter)
                    self._dataframe = pd.DataFrame(
                        list(queryset), columns=columns)
                self._dataframe.rename(columns=self.columns, inplace=True)
                self._dataframe.fillna(value=np.nan, inplace=True)
                for column in list(self._dataframe.select_dtypes(
                        include=['datetime64[ns, UTC]']).columns):
                    self._dataframe[column] = self._dataframe[
                        column].astype('datetime64[ns]')
            if self.drop_sys_columns:
                self._dataframe = self._dataframe.drop(
                    self.edc_sys_columns, axis=1)
        return self._dataframe

    def get_column_value(self, model_obj=None, column_name=None, lookup=None):
        """Returns the column value.
        """
        lookups = {column_name: lookup} if LOOKUP_SEP in lookup else None
        value_getter = self.value_getter_cls(
            field_name=column_name,
            model_obj=model_obj,
            lookups=lookups)
        return value_getter.value

    @property
    def model_cls(self):
        return django_apps.get_model(self.model)

    @property
    def has_encrypted_fields(self):
        """Returns True if at least one field uses encryption.
        """
        for field in self.model_cls._meta.fields:
            if hasattr(field, 'field_cryptor'):
                return True
        return False

    @property
    def encrypted_columns(self):
        """Return a list of column names that use encryption.
        """
        if not self._encrypted_columns:
            self._encrypted_columns = ['identity_or_pk']
            for field in self.model_cls._meta.fields:
                if hasattr(field, 'field_cryptor'):
                    self._encrypted_columns.append(field.name)
            self._encrypted_columns = list(set(self._encrypted_columns))
            self._encrypted_columns.sort()
        return self._encrypted_columns

    @property
    def columns(self):
        """Return a dictionary of column names.
        """
        if not self._columns:
            columns = list(self.queryset[0].__dict__.keys())
            for name in self.sys_field_names:
                try:
                    columns.remove(name)
                except ValueError:
                    pass
            columns = dict(zip(columns, columns))
            if self.add_columns_for in columns or f'{self.add_columns_for}_id' in columns:
                if self.add_columns_for.endswith('_visit'):
                    columns.update({
                        f'{self.add_columns_for}__appointment__visit_code': 'visit_code'})
                    try:
                        del columns['subject_identifier']
                    except KeyError:
                        columns.update({
                            f'{self.add_columns_for}__appointment__subject_identifier':
                            'subject_identifier'})
            self._columns = columns
        return self._columns



import numpy as np
import pandas as pd

from edc_constants.constants import YES

from ..mappings import yes_no
from .table import Table


class Requisition(Table):

    default_columns = ['subject_visit_id', 'requisition_identifier',
                       'panel_name', 'requisition_datetime',
                       'is_drawn', 'reason_not_drawn', 'drawn_datetime',
                       'specimen_type', 'study_site_name']

    def prepare(self, df_visit=None, **kwargs):
        self.dataframe['requisition_datetime'] = self.helper.to_local_datetime(
            self.dataframe['requisition_datetime'])
        self.dataframe['drawn_datetime'] = self.helper.to_local_datetime(
            self.dataframe['drawn_datetime'])
        self.dataframe['received_datetime'] = self.helper.to_local_datetime(
            self.dataframe['received_datetime'])
        self.dataframe['received'] = self.dataframe['received'].map(
            yes_no, na_action='ignore')
        self.dataframe['processed'] = self.dataframe['processed'].map(
            yes_no, na_action='ignore')
        self.dataframe['packed'] = self.dataframe['packed'].map(
            yes_no, na_action='ignore')
        self.dataframe['shipped'] = self.dataframe['shipped'].map(
            yes_no, na_action='ignore')
        self.dataframe = pd.merge(
            self.dataframe, df_visit, on='subject_visit_id', how='left')
        self.dataframe = self.dataframe.drop(['subject_visit_id'], axis=1)
        # fix reason_not_drawn
        is_blank = (self.dataframe.reason_not_drawn.notnull()) & (
            self.dataframe.is_drawn == YES)
        self.dataframe.loc[is_blank, 'reason_not_drawn'] = np.nan
        self.dataframe = self.dataframe[self.reordered_columns]

    @property
    def reordered_columns(self):
        columns = [col for col in list(self.dataframe.columns) if col not in [
            'subject_identifier', 'gender', 'dob',
            'visit_code', 'visit_datetime', 'visit_date']]
        columns.insert(0, 'visit_code')
        columns.insert(0, 'visit_date')
        columns.insert(0, 'visit_datetime')
        columns.insert(0, 'dob')
        columns.insert(0, 'gender')
        columns.insert(0, 'subject_identifier')
        return columns



import csv
import os

from django.apps import apps as django_apps
from django.test import TestCase, tag
from tempfile import mkdtemp

from ..csv_exporters import CsvModelExporter
from ..model_to_dataframe import ModelToDataframe
from .models import Crf, CrfEncrypted, SubjectVisit
from .helper import Helper

app_config = django_apps.get_app_config('edc_pdutils')


class TestExport(TestCase):

    helper = Helper()

    def setUp(self):
        self.path = mkdtemp()
        for i in range(0, 5):
            self.helper.create_crf(i)
        self.subject_visit = SubjectVisit.objects.all()[0]

    def test_none(self):
        Crf.objects.all().delete()
        model = 'edc_pdutils.crf'
        m = ModelToDataframe(model=model, add_columns_for='subject_visit')
        self.assertEqual(len(m.dataframe.index), 0)

    def test_records(self):
        model = 'edc_pdutils.crf'
        m = ModelToDataframe(model=model, add_columns_for='subject_visit')
        self.assertEqual(len(m.dataframe.index), 5)
        model = 'edc_pdutils.crfone'
        m = ModelToDataframe(model=model, add_columns_for='subject_visit')
        self.assertEqual(len(m.dataframe.index), 5)

    def test_records_as_qs(self):
        m = ModelToDataframe(queryset=Crf.objects.all(),
                             add_columns_for='subject_visit')
        self.assertEqual(len(m.dataframe.index), 5)

    def test_columns(self):
        model = 'edc_pdutils.crf'
        m = ModelToDataframe(model=model, add_columns_for='subject_visit')
        self.assertEqual(len(list(m.dataframe.columns)), 18)

    def test_values(self):
        model = 'edc_pdutils.crf'
        m = ModelToDataframe(model=model, add_columns_for='subject_visit')
        df = m.dataframe
        df.sort_values(['subject_identifier'], inplace=True)
        for i in range(0, 5):
            self.assertEqual(df.subject_identifier.iloc[i], f'12345{i}')
            self.assertEqual(df.visit_code.iloc[i], f'{i}000')

    def test_encrypted_none(self):
        model = 'edc_pdutils.crfencrypted'
        m = ModelToDataframe(model=model, add_columns_for='subject_visit')
        self.assertEqual(len(m.dataframe.index), 0)

    def test_encrypted_records(self):
        CrfEncrypted.objects.create(
            subject_visit=self.subject_visit,
            encrypted1=f'encrypted1')
        model = 'edc_pdutils.crfencrypted'
        m = ModelToDataframe(model=model, add_columns_for='subject_visit')
        self.assertEqual(len(m.dataframe.index), 1)

    def test_encrypted_records_as_qs(self):
        CrfEncrypted.objects.create(
            subject_visit=self.subject_visit,
            encrypted1=f'encrypted1')
        m = ModelToDataframe(
            queryset=CrfEncrypted.objects.all(), add_columns_for='subject_visit')
        self.assertEqual(len(m.dataframe.index), 1)

    def test_encrypted_to_csv_from_qs(self):
        CrfEncrypted.objects.create(
            subject_visit=self.subject_visit,
            encrypted1=f'encrypted1')
        model_exporter = CsvModelExporter(
            queryset=CrfEncrypted.objects.all(),
            add_columns_for='subject_visit',
            export_folder=self.path)
        model_exporter.to_csv()

    def test_encrypted_to_csv_from_model(self):
        CrfEncrypted.objects.create(
            subject_visit=self.subject_visit,
            encrypted1=f'encrypted1')
        model_exporter = CsvModelExporter(
            model='edc_pdutils.CrfEncrypted',
            add_columns_for='subject_visit',
            export_folder=self.path)
        model_exporter.to_csv()

    def test_records_to_csv_from_qs(self):
        model_exporter = CsvModelExporter(
            queryset=Crf.objects.all(),
            add_columns_for='subject_visit',
            export_folder=self.path)
        model_exporter.to_csv()

    def test_records_to_csv_from_model(self):
        model_exporter = CsvModelExporter(
            model='edc_pdutils.crf',
            add_columns_for='subject_visit',
            sort_by=['subject_identifier'],
            export_folder=self.path)
        path = model_exporter.to_csv()
        with open(path, 'r') as f:
            csv_reader = csv.DictReader(f, delimiter='|')
            rows = [row for row in enumerate(csv_reader)]
        self.assertEqual(len(rows), 5)
        for i in range(0, 5):
            self.assertEqual(rows[i][1].get('subject_identifier'), f'12345{i}')
            self.assertEqual(rows[i][1].get('visit_code'), f'{i}000')



import sys
import pandas as pd

from .table import Table


class Consent(Table):

    default_columns = ['subject_identifier', 'gender',
                       'dob', 'consent_datetime', 'study_site']

    def prepare(self, **kwargs):
        self.dataframe['dob'] = self.helper.date_to_local_datetime(
            self.dataframe['dob'])
        self.dataframe['consent_datetime'] = self.helper.to_local_datetime(
            self.dataframe['consent_datetime'])
        self.dataframe = self.dataframe.sort_values(['subject_identifier'])
        self.dataframe = self.dataframe.drop_duplicates()
        self.dataframe = self.dataframe.groupby(
            ['subject_identifier', 'consent_datetime']).min().reset_index()
        self.dataframe['duplicate'] = self.dataframe.duplicated(self.dataframe)
        if len(self.dataframe[self.dataframe['duplicate']]) > 0:
            self.dataframe = self.deduplicate()
        self.dataframe['consent_date'] = self.dataframe['consent_datetime'].dt.normalize()
        self.columns.append('consent_date')

    def deduplicate(self):
        sys.stderr.write('Consent has duplicates!')
        sys.stderr.write(self.dataframe[self.dataframe['duplicate']])

    def read_csv(self, **kwargs):
        super().read_csv(**kwargs)
        self.dataframe['consent_datetime'] = self.dataframe['consent_datetime'].astype(
            'datetime64[ns]')
        self.dataframe['consent_date'] = self.dataframe['consent_date'].astype(
            'datetime64[ns]')



import os
import shutil

from django.contrib.auth.models import User
from edc_registration.models import RegisteredSubject
from django.test import TestCase, tag
from django.test.utils import override_settings
from tempfile import mkdtemp

from ..archive_exporter import ArchiveExporter, NothingToExport


@override_settings(EXPORT_FOLDER=mkdtemp())
class TestArchiveExporter(TestCase):

    def setUp(self):

        User.objects.create(username='erikvw')
        RegisteredSubject.objects.create(subject_identifier='12345')
        self.models = [
            'auth.user',
            'edc_registration.registeredsubject']

    @tag('1')
    def test_request_archive(self):

        exporter = ArchiveExporter()
        history = exporter.export_to_archive(models=self.models)
        folder = mkdtemp()
        shutil.unpack_archive(
            history.archive_filename, folder, 'zip')
        filenames = os.listdir(folder)
        self.assertGreater(
            len([f for f in filenames]), 0)

    def test_request_archive_filename_exists(self):

        exporter = ArchiveExporter()
        history = exporter.export_to_archive(
            models=self.models, user='erikvw')
        filename = history.archive_filename
        self.assertIsNotNone(filename)
        self.assertTrue(
            os.path.exists(filename),
            msg=f'file \'{filename}\' does not exist')

    def test_requested_with_invalid_table(self):
        models = [
            'auth.blah',
            'edc_registration.registeredsubject']
        exporter = ArchiveExporter()
        self.assertRaises(
            LookupError, exporter.export_to_archive,
            models=models, user='erikvw')

    def test_requested_with_nothing(self):
        models = []
        exporter = ArchiveExporter()
        self.assertRaises(
            NothingToExport, exporter.export_to_archive,
            models=models, user='erikvw')



import pandas as pd

from copy import copy
from edc_base import DEFAULT_BASE_FIELDS

from ..mappings import yes_no
from .table import Table


class Aliquot(Table):

    default_columns = None
    requisition_columns = [
        'requisition_identifier',
        'panel_name', 'requisition_datetime', 'drawn_datetime',
        'gender', 'dob', 'study_site', 'study_site_name',
        'visit_code', 'visit_datetime', 'visit_date']

    def prepare(self, df_requisition=None, requisition_columns=None, **kwargs):
        self.requisition_columns = requisition_columns or self.requisition_columns
        self.dataframe['aliquot_datetime'] = self.helper.to_local_datetime(
            self.dataframe['aliquot_datetime'])
        self.dataframe['aliquot_date'] = (
            self.dataframe['aliquot_datetime'].dt.normalize())

        # drop sys and other unwanted columns
        base_cols = copy(DEFAULT_BASE_FIELDS)
        base_cols.pop(base_cols.index('hostname_created'))
        base_cols = base_cols + ['slug', 'comment', 'shipped']
        self.dataframe = self.dataframe.drop(base_cols, axis=1)

        # add column to indicate missing requisitions, remap columns
        self.dataframe['missing_requisition'] = self.missing_requisition(
            df_requisition)
        self.dataframe['is_primary'] = self.dataframe['is_primary'].map(
            yes_no, na_action='ignore')
        self.dataframe['medium'] = self.dataframe['medium'].str.lower()
        self.dataframe['parent_identifier'] = self.dataframe.apply(
            lambda row: self.fix_parent_identifier(row['parent_identifier']), axis=1)
        self.dataframe = pd.merge(
            self.dataframe, df_requisition[self.requisition_columns],
            on='requisition_identifier', how='left')
        self.dataframe = self.dataframe[self.reordered_columns]

    @property
    def reordered_columns(self):
        columns = [col for col in list(self.dataframe.columns) if col not in [
            'hostname_created', 'subject_identifier', 'gender', 'dob',
            'visit_code', 'visit_datetime', 'visit_date']]
        columns.insert(0, 'visit_code')
        columns.insert(0, 'visit_date')
        columns.insert(0, 'visit_datetime')
        columns.insert(0, 'dob')
        columns.insert(0, 'gender')
        columns.insert(0, 'subject_identifier')
        columns.append('hostname_created')
        return columns

    def fix_parent_identifier(self, value):
        if value:
            value = value.replace('(', '').replace(')', '').replace(
                '\'', '').replace(',', '')
        return value

    def missing_requisition(self, df_requisition=None):
        """Returns a Series with True for rows missing a requisition.
        """
        missing = pd.merge(self.dataframe, df_requisition[self.requisition_columns],
                           on='requisition_identifier', how='left')
        return missing['panel_name'].isnull()



import os

from django.apps import apps as django_apps
from django.conf import settings
from django.db import models
from edc_base.model_managers import HistoricalRecords
from edc_base.model_mixins import BaseUuidModel

from .choices import EXPORT_FORMATS
from .constants import CSV


class DataRequest(BaseUuidModel):

    requested = models.TextField(
        help_text='List one table per line, no commas. Use "label lower" format.')

    archive_filename = models.CharField(max_length=200, null=True)

    decrypt = models.BooleanField(default=False)

    export_format = models.CharField(
        max_length=25,
        choices=EXPORT_FORMATS,
        default=CSV)

    exported = models.BooleanField(default=False)

    exported_datetime = models.DateTimeField(null=True)

    history = HistoricalRecords()

    def __str__(self):
        if self.archive_filename:
            return os.path.basename(self.archive_filename)
        else:
            return 'pending'

    @property
    def requested_as_list(self):
        """Returns `requested` as a list.

        Validates each item to be a model name.
        """
        requested = self.requested.split('\n')
        requested = [x.strip() for x in requested if x.strip()]
        for model in requested:
            django_apps.get_model(model)
        return requested


if settings.APP_NAME == 'edc_pdutils':
    from .tests import models



import sys
import pandas as pd
import numpy as np

from django.conf import settings
from django.db import connection
from uuid import UUID

from .dialects import MysqlDialect


class DatabaseNameError(Exception):
    pass


class Database:

    dialect_cls = MysqlDialect
    lowercase_columns = True
    DATABASES_NAME = 'default'

    def __init__(self, **kwargs):
        self._database = None
        self._tables = pd.DataFrame()
        self.dialect = self.dialect_cls(dbname=self.database)

    @property
    def database(self):
        """Returns the database name.
        """
        if not self._database:
            filename = settings.DATABASES.get(self.DATABASES_NAME).get(
                'OPTIONS').get('read_default_file')
            if 'test' in sys.argv:
                self._database = connection.settings_dict['TEST']['NAME']
            else:
                self._database = connection.settings_dict['NAME']
                if not self._database:
                    with open(filename, 'r') as f:
                        for line in [line for line in f if '#' not in line]:
                            if 'database' in line:
                                self._database = line.split('=')[1].strip()
            if not self._database:
                raise DatabaseNameError(
                    f'Unable to determine the DB name from settings.DATABASES. '
                    f'Got NAME={self.DATABASES_NAME}, read_default_file={filename}.')
        return self._database

    def read_sql(self, sql, params=None):
        """Returns a dataframe. A simple wrapper for pd.read_sql().
        """
        return pd.read_sql(sql, connection, params=params)

    def show_databases(self):
        """Returns a dataframe of database names in the schema.
        """
        sql, params = self.dialect.show_databases()
        return self.read_sql(sql, params=params)

    def select_table(self, table_name=None, lowercase_columns=None, uuid_columns=None,
                     limit=None):
        """Returns a dataframe of a table.

        Note: UUID columns are stored as strings (uuid.hex) and need
        to be converted from string to UUID if to match the
        rendering of the same column by a Django model class.
        """
        uuid_columns = uuid_columns or []
        lowercase_columns = lowercase_columns or self.lowercase_columns
        sql, params = self.dialect.select_table(table_name)
        if limit:
            sql = f'{sql} LIMIT {int(limit)}'
        df = self.read_sql(sql, params=params)
        if lowercase_columns:
            columns = {col: col.lower() for col in list(df.columns)}
            df.rename(columns=columns, inplace=True)
        for col in uuid_columns:
            df[col] = df.apply(
                lambda row: str(
                    UUID(row[col])) if row[col] else np.nan,
                axis=1)
        return df

    def show_tables(self, app_label=None):
        """Returns a dataframe of table names in the schema.
        """
        sql, params = self.dialect.show_tables(app_label)
        return self.read_sql(sql, params=params)

    def show_tables_with_columns(self, app_label=None, column_names=None):
        """Returns a dataframe of table names in the schema
        that have a column in column_names.
        """
        sql, params = self.dialect.show_tables_with_columns(
            app_label, column_names)
        return self.read_sql(sql, params=params)

    def show_tables_without_columns(self, app_label=None, column_names=None):
        """Returns a dataframe of table names in the schema.
        that DO NOT have a column in column_names.
        """
        sql, params = self.dialect.show_tables_without_columns(
            app_label, column_names)
        return self.read_sql(sql, params=params)

    def show_inline_tables(self, referenced_table_name=None):
        sql, params = self.dialect.show_inline_tables(referenced_table_name)
        return self.read_sql(sql, params=params)



from .model_to_dataframe import ModelToDataframe


class SubjectModelToDataframe(ModelToDataframe):

    columns = {k: k for k in ['subject_identifier', 'gender', 'dob']}

    def __init__(self, columns=None, **kwargs):
        if columns:
            self.columns = {k: k for k in columns}
        super().__init__(**kwargs)



import os
import sys

from django.core.management.color import color_style
from edc_base import get_utcnow


style = color_style()


class CsvExporterExportFolder(Exception):
    pass


class CsvExporterFileExists(Exception):
    pass


class CsvExporter:

    delimiter = '|'
    encoding = 'utf-8'
    export_folder = None
    index = False
    file_exists_ok = False
    date_format = None
    sort_by = None

    def __init__(self, data_label=None, sort_by=None, export_folder=None,
                 delimiter=None, date_format=None, index=None, **kwargs):
        self.delimiter = delimiter or self.delimiter
        self.date_format = date_format or self.date_format
        self.index = index or self.index
        self.sort_by = sort_by or self.sort_by
        self.export_folder = export_folder or self.export_folder
        if not os.path.exists(self.export_folder):
            raise CsvExporterExportFolder(
                f'Invalid export folder. Got {self.export_folder}')
        self.data_label = data_label

    def to_csv(self, dataframe=None, export_folder=None):
        """Returns the full path of the written CSV file if the
        dataframe is exported otherwise None.

        Note: You could also just do:
            >>> dataframe.to_csv(path_or_buf=path, **self.csv_options)
            to suppress stdout messages.
        """
        path = None
        sys.stdout.write(self.data_label + '\r')
        if export_folder:
            self.export_folder = export_folder
        if not dataframe.empty:
            path = self.path
            if self.sort_by:
                dataframe.sort_values(self.sort_by, inplace=True)
            sys.stdout.write(f'( ) {self.data_label} ...     \r')
            dataframe.to_csv(path_or_buf=path, **self.csv_options)
            recs = len(dataframe)
            sys.stdout.write(
                f'({style.SUCCESS("*")}) {self.data_label} {recs}       \n')
        else:
            sys.stdout.write(f'(?) {self.data_label} empty  \n')
        return path

    @property
    def csv_options(self):
        """Returns default options for dataframe.to_csv().
        """
        return dict(
            index=self.index,
            encoding=self.encoding,
            sep=self.delimiter,
            date_format=self.date_format)

    @property
    def path(self):
        """Returns a full path and filename.
        """
        path = os.path.join(self.export_folder, self.filename)
        if os.path.exists(path) and not self.file_exists_ok:
            raise CsvExporterFileExists(
                f'File \'{path}\' exists! Not exporting {self.data_label}.\n')
        return path

    @property
    def filename(self):
        """Returns a CSV filename based on the timestamp.
        """
        dt = get_utcnow()
        prefix = self.data_label.replace('-', '_')
        formatted_date = dt.strftime('%Y%m%d%H%M%S')
        return f'{prefix}_{formatted_date}.csv'



import os
import shutil
import sys

from django.apps import apps as django_apps
from django.conf import settings
from edc_base.utils import get_utcnow
from tempfile import mkdtemp

from .csv_exporters import CsvModelExporter


class NothingToExport(Exception):
    pass


class ArchiveExporter:

    """Exports a list of models to individual CSV files and
    adds each to a single zip archive.
    """

    date_format = '%Y%m%d%H%M%S'
    history_model = 'edc_pdutils.datarequest'
    csv_exporter_cls = CsvModelExporter

    def __init__(self, export_folder=None, date_format=None):
        self.date_format = date_format or self.date_format
        self.export_folder = export_folder or settings.EXPORT_FOLDER
        self.history_model_cls = django_apps.get_model(self.history_model)

    def export_to_archive(self, models=None, decrypt=None, user=None, **kwargs):
        """Returns a history model instance after exporting
         models to a single zip archive file.

        models: a list of model names in label_lower format.
        """
        history = None
        exported = []
        tmp_folder = mkdtemp()
        user = user or 'unknown_user'
        for model in models:
            print(model)
            csv_exporter = self.csv_exporter_cls(
                model=model,
                export_folder=tmp_folder,
                decrypt=decrypt, **kwargs)
            exported.append(csv_exporter.to_csv())
        if not exported:
            raise NothingToExport(
                f'Nothing exported. Got models={models}.')
        else:
            archive_filename = self._archive(tmp_folder=tmp_folder, user=user)
            exported_datetime = get_utcnow()
            history = self.history_model_cls.objects.create(
                requested='\n'.join(models),
                decrypt=False if decrypt is None else decrypt,
                exported_datetime=exported_datetime,
                archive_filename=archive_filename,
                user_created=user,
                exported=True)
            sys.stdout.write(
                f'\nExported archive to {history.archive_filename}.\n')
        return history

    def _archive(self, tmp_folder=None, exported_datetime=None, user=None):
        """Returns the archive zip filename after calling make_archive.
        """
        exported_datetime = exported_datetime or get_utcnow()
        formatted_date = exported_datetime.strftime(self.date_format)
        return shutil.make_archive(
            os.path.join(self.export_folder, f'{user}_{formatted_date}'), 'zip', tmp_folder)



from django.contrib.admin import AdminSite as DjangoAdminSite


class AdminSite(DjangoAdminSite):

    site_title = 'EDC Pandas Utils'
    site_header = 'EDC Pandas Utils'
    index_title = 'EDC Pandas Utils'
    site_url = '/administration/'


edc_pdutils_admin = AdminSite(name='edc_pdutils_admin')



import os

from django.apps import AppConfig as DjangoApponfig
from django.conf import settings


class AppConfig(DjangoApponfig):
    name = 'edc_pdutils'
    verbose_name = 'Edc Export'
    export_folder = os.path.join(settings.MEDIA_ROOT, 'edc_pdutils', 'export')

    def ready(self):
        os.makedirs(self.export_folder, exist_ok=True)



import uuid

from edc_appointment.models import Appointment
from edc_base.utils import get_utcnow
from edc_registration.models import RegisteredSubject

from .models import ListModel, SubjectVisit, Crf
from .models import CrfTwo, CrfOne, CrfThree, CrfInline


class Helper:

    def create_crf(self, i=None):
        i = i or 0
        subject_identifier = f'12345{i}'
        visit_code = f'{i}000'
        RegisteredSubject.objects.create(
            subject_identifier=subject_identifier)
        appointment = Appointment.objects.create(
            subject_identifier=subject_identifier,
            visit_code=visit_code,
            appt_datetime=get_utcnow())
        self.thing_one = ListModel.objects.create(
            name=f'thing_one_{i}', short_name=f'thing_one_{i}')
        self.thing_two = ListModel.objects.create(
            name=f'thing_two_{i}', short_name=f'thing_two_{i}')
        self.subject_visit = SubjectVisit.objects.create(
            appointment=appointment,
            subject_identifier=subject_identifier,
            report_datetime=get_utcnow())
        Crf.objects.create(
            subject_visit=self.subject_visit,
            char1=f'char{i}',
            date1=get_utcnow(),
            int1=i,
            uuid1=uuid.uuid4())
        crf_one = CrfOne.objects.create(
            subject_visit=self.subject_visit,
            dte=get_utcnow())
        crf_two = CrfTwo.objects.create(
            subject_visit=self.subject_visit,
            dte=get_utcnow())
        CrfThree.objects.create(
            subject_visit=self.subject_visit,
            UPPERCASE=get_utcnow())
        CrfInline.objects.create(
            crf_one=crf_one,
            crf_two=crf_two,
            dte=get_utcnow())



from django.test import TestCase, tag
from edc_pdutils.dialects.crf_dialect import CrfDialect


class DummyDfHandler:
    visit_tbl = 'edc_pdutils_subjectvisit'
    appointment_tbl = 'edc_appointment_appointment'
    registered_subject_tbl = 'edc_registration_registeredsubject'

    def __init__(self, visit_column=None):
        self.visit_column = visit_column or 'subject_visit_id'


class TestDialects(TestCase):

    def test_init(self):
        obj = DummyDfHandler()
        crf_dialect = CrfDialect(obj=obj)
        self.assertTrue(
            crf_dialect.select_visit_and_related)
        self.assertIn(
            'SELECT',
            crf_dialect.select_visit_and_related[0])
        self.assertIn(
            obj.visit_column,
            crf_dialect.select_visit_and_related[0])
        self.assertIn(
            obj.appointment_tbl,
            crf_dialect.select_visit_and_related[0])
        self.assertIn(
            obj.registered_subject_tbl,
            crf_dialect.select_visit_and_related[0])
        self.assertIn(
            obj.visit_tbl,
            crf_dialect.select_visit_and_related[0])

    def test_uses_df_handlers_visit_column(self):
        obj = DummyDfHandler(visit_column='erik_visit_id')
        crf_dialect = CrfDialect(obj=obj)
        self.assertTrue(
            crf_dialect.select_visit_and_related)
        self.assertIn(
            'SELECT',
            crf_dialect.select_visit_and_related[0])
        self.assertIn(
            obj.visit_column,
            crf_dialect.select_visit_and_related[0])



import csv
import sys

from django.apps import apps as django_apps
from django.test import TestCase, tag
from tempfile import mkdtemp

from ..csv_exporters import CsvCrfTablesExporter, CsvCrfInlineTablesExporter
from ..df_handlers import CrfDfHandler
from .helper import Helper

app_config = django_apps.get_app_config('edc_pdutils')


class TestExport(TestCase):

    helper = Helper()

    def setUp(self):
        self.path = mkdtemp()
        for i in range(0, 5):
            self.helper.create_crf(i)

    @tag('1')
    def test_crf_tables_to_csv_from_app_label_with_columns(self):

        class MyCrfDfHandler(CrfDfHandler):
            visit_column = 'subject_visit_id'
            visit_tbl = 'edc_pdutils_subjectvisit'
            registered_subject_tbl = 'edc_registration_registeredsubject'
            appointment_tbl = 'edc_appointment_appointment'

        class MyCsvCrfTablesExporter(CsvCrfTablesExporter):
            visit_column = 'subject_visit_id'
            df_handler_cls = MyCrfDfHandler
            app_label = 'edc_pdutils'
            export_folder = self.path

        sys.stdout.write('\n')
        exporter = MyCsvCrfTablesExporter()
        exporter.to_csv()
        self.assertGreater(len(exporter.exported_paths), 0)
        for path in exporter.exported_paths.values():
            with open(path, 'r') as f:
                csv_reader = csv.DictReader(f, delimiter='|')
                rows = [row for row in enumerate(csv_reader)]
            self.assertGreater(len(rows), 0)

    def test_export_inlines(self):

        class MyDfHandler(CrfDfHandler):
            visit_column = 'subject_visit_id'
            visit_tbl = 'edc_pdutils_subjectvisit'
            registered_subject_tbl = 'edc_registration_registeredsubject'
            appointment_tbl = 'edc_appointment_appointment'

        class MyCsvCrfInlineTablesExporter(CsvCrfInlineTablesExporter):
            visit_column = 'subject_visit_id'
            df_handler_cls = MyDfHandler
            app_label = 'edc_pdutils'
            export_folder = self.path

        sys.stdout.write('\n')
        exporter = MyCsvCrfInlineTablesExporter()
        exporter.to_csv()
        self.assertGreater(len(exporter.exported_paths), 0)
        for path in exporter.exported_paths.values():
            with open(path, 'r') as f:
                csv_reader = csv.DictReader(f, delimiter='|')
                rows = [row for row in enumerate(csv_reader)]
            self.assertGreater(len(rows), 0)



from django.conf import settings

from .constants import CSV


APP_LABELS = (
    ((x, x) for x in [app_label.split('.')[0]
                      for app_label in settings.INSTALLED_APPS]),
)

EXPORT_FORMATS = (
    (CSV, 'CSV'),
)



import os
import pandas as pd

from datetime import datetime

from ..helper import Helper as BaseHelper
from .table_to_dataframe import TableToDataframe


class Helper(BaseHelper):

    to_dataframe_cls = TableToDataframe

    def get_raw_dataframe(self, table_name=None, drop_sys_columns=None,
                          csv_filename=None, path=None, timestamp=None):
        """Returns a dataframe.
        """
        if csv_filename:
            timestamp = timestamp or datetime.today().strftime('%Y%m%d%H%M%S')
            csv_filename = f'{table_name}_{timestamp}.csv'
            df_raw = self.to_dataframe_cls(
                table_name=table_name,
                drop_sys_columns=drop_sys_columns).dataframe
            df_raw.to_csv(os.path.join(path, csv_filename), index=False)
            print(os.path.join(path, csv_filename))
        else:
            df_raw = pd.read_csv(
                os.path.join(path, csv_filename), low_memory=False)
        return df_raw



import pandas as pd

from .table import Table


class Visit(Table):

    default_columns = ['id', 'subject_identifier',
                       'report_datetime', 'visit_code']

    def prepare(self, df_subjects=None, **kwargs):
        self.dataframe = self.dataframe.rename(
            columns={'id': 'subject_visit_id', 'report_datetime': 'visit_datetime'})
        self.columns[self.columns.index('report_datetime')] = [
            'visit_datetime']
        self.columns[self.columns.index('id')] = [
            'subject_visit_id']
        self.dataframe['visit_datetime'] = self.helper.to_local_datetime(
            self.dataframe['visit_datetime'])
        self.dataframe['visit_date'] = self.dataframe['visit_datetime'].dt.normalize()
        self.columns.append('visit_date')
        self.dataframe = pd.merge(self.dataframe, df_subjects,
                                  on='subject_identifier', how='left')

    def read_csv(self, **kwargs):
        super().read_csv(**kwargs)
        self.dataframe['visit_datetime'] = self.dataframe['visit_datetime'].astype(
            'datetime64[ns]')
        self.dataframe['visit_date'] = self.dataframe['visit_date'].astype(
            'datetime64[ns]')
        self.dataframe['consent_datetime'] = self.dataframe['consent_datetime'].astype(
            'datetime64[ns]')
        self.dataframe['consent_date'] = self.dataframe['consent_date'].astype(
            'datetime64[ns]')



import pandas as pd

from edc_pdutils.model_to_dataframe.subject_model_to_dataframe import SubjectModelToDataframe


def missing_subject_identifiers(self, model=None, subject_identifiers=None,
                                remove_uuids=None, verbose=None):
    """Returns a series of subject_identifiers not in model.

    For example:
        missing_subject_identifiers(
            model='edc_registration.registeredsubject',
            subject_identifiers=[a list of subject identifiers])
    """
    # convert list of subject identifiers to a dataframe
    df_subject_identifiers = pd.DataFrame(
        subject_identifiers, columns=['subject_identifier'])
    df_subject_identifiers['identifier'] = df_subject_identifiers['subject_identifier']
    df_subject_identifiers = df_subject_identifiers.set_index(
        'subject_identifier')
    if verbose:
        df_subject_identifiers.head()

    # load model into dataframe
    df_subject = SubjectModelToDataframe(model=model).dataframe
    # drop duplicates
    df_subject = df_subject.drop_duplicates()
    # remove subject identifier as UUID
    if remove_uuids:
        df_subject = df_subject[
            df_subject['subject_identifier'].str.len() != 32]
    # set index to subject_identifier
    df_subject = df_subject.set_index('subject_identifier')
    # filter df of subject identifiers to leave only those not in model
    df_missing = df_subject_identifiers[-df_subject_identifiers['identifier'].isin(
        df_subject.index)]

    if len(df_missing.index) > 0 and verbose:
        print(
            f'There are {len(df_missing["identifier"])} subject identifiers '
            f'missing from {model}.')
    return df_missing['identifier']



from django.core.management.color import color_style

from ..model_to_dataframe import ModelToDataframe
from .csv_exporter import CsvExporter


style = color_style()


class CsvModelExporter:

    df_maker_cls = ModelToDataframe
    csv_exporter_cls = CsvExporter

    def __init__(self, model=None, queryset=None, decrypt=None, sort_by=None,
                 **kwargs):
        self.model = model or queryset.model._meta.label_lower
        self.df_maker = self.df_maker_cls(
            model=model, queryset=queryset, decrypt=decrypt, **kwargs)
        self.csv_exporter = self.csv_exporter_cls(
            data_label=self.model, sort_by=sort_by, **kwargs)

    def to_csv(self):
        return self.csv_exporter.to_csv(self.df_maker.dataframe)



import csv
import os
import sys

from django.apps import apps as django_apps
from django.test import TestCase, tag

from ..csv_exporters import CsvTablesExporter
from .helper import Helper

app_config = django_apps.get_app_config('edc_pdutils')


class TestExport(TestCase):

    path = app_config.export_folder
    helper = Helper()

    def setUp(self):
        for i in range(0, 5):
            self.helper.create_crf(i)

    def tearDown(self):
        """Remove .csv files created in tests.
        """
        super().tearDown()
        if 'edc_pdutils' not in self.path:
            raise ValueError(f'Invalid path in test. Got {self.path}')
        files = os.listdir(self.path)
        for file in files:
            if '.csv' in file:
                file = os.path.join(self.path, file)
                os.remove(file)

    def test_tables_to_csv_lower_columns(self):
        sys.stdout.write('\n')
        tables_exporter = CsvTablesExporter(app_label='edc_pdutils')
        for path in tables_exporter.exported_paths.values():
            with open(path, 'r') as f:
                csv_reader = csv.DictReader(f, delimiter='|')
                for row in csv_reader:
                    for field in row:
                        self.assertEqual(field.lower(), field)
                    break

    def test_tables_to_csv_from_app_label(self):
        sys.stdout.write('\n')
        tables_exporter = CsvTablesExporter(app_label='edc_pdutils')
        for path in tables_exporter.exported_paths.values():
            with open(path, 'r') as f:
                csv_reader = csv.DictReader(f, delimiter='|')
                rows = [row for row in enumerate(csv_reader)]
            self.assertGreater(len(rows), 0)

    def test_tables_to_csv_from_app_label_exclude_history(self):

        class MyCsvTablesExporter(CsvTablesExporter):
            exclude_history_tables = True

        sys.stdout.write('\n')
        tables_exporter = MyCsvTablesExporter(app_label='bcpp_clinic')
        for path in tables_exporter.exported_paths:
            self.assertNotIn('history', path)



import string

from django.core.exceptions import ValidationError
from django.db.models.constants import LOOKUP_SEP
from django_crypto_fields.fields import BaseField as BaseEncryptedField


class ValueGetterUnknownField(ValidationError):
    pass


class ValueGetterInvalidLookup(ValidationError):
    pass


class ValueGetter:

    """A class to retrieve a field value from a django models
    instance.

    Keyword Args:
        * field_name: model instance field name
        * model_obj: model instance
        * lookups: a dictionary of {field_name: lookup, ...} where
            lookup is a django-style lookup string
            e.g. 'subject_visit__appointment__visit_code'.
        * encrypt: replaces encrypted values with `encrypted_label`.
            (Default: True). If False will decrypt the value.
        * additional_values: a class to provide additional attrs
            to get if field_name is not on the model_obj and
            not a lookup.
    """

    encrypted_label = '<encrypted>'
    m2m_delimiter = ';'

    def __init__(self, field_name=None, model_obj=None,
                 lookups=None, encrypt=None, additional_values=None):
        self._value = None
        self.additional_values = additional_values
        self.field_name = field_name
        self.model_obj = model_obj
        self.lookups = lookups or {}
        self.model_cls = self.model_obj.__class__
        self.encrypt = True if encrypt is None else encrypt

    @property
    def value(self):
        if not self._value:
            try:
                self._value = self.get_field_value(
                    self.model_obj, self.field_name)
            except ValueGetterUnknownField as e:
                try:
                    self._value = getattr(self.model_obj, e.code)
                except AttributeError:
                    if self.additional_values:
                        self._value = getattr(self.additional_values, e.code)
                    else:
                        raise ValueGetterUnknownField(e)
            if self._value is None:
                self._value = ''
            self._value = self.strip_value(self._value)
        return self._value

    def get_field_value(self, model_obj=None, field_name=None):
        """Returns a field value.
        """
        value = ''
        for f in model_obj.__class__._meta.fields:
            if (f.name == field_name and issubclass(f.__class__, BaseEncryptedField)
                    and self.encrypt):
                value = self.encrypted_label
        if value != self.encrypted_label:
            try:
                value = self.getattr(model_obj, field_name)
            except AttributeError:
                if field_name in self.lookups:
                    value = self.get_lookup_value(
                        model_obj=model_obj,
                        field_name=field_name)
                elif field_name in self.m2m_field_names:
                    value = self.get_m2m_value(model_obj, field_name)
                else:
                    raise ValueGetterUnknownField(
                        f'Unknown field name. Got {field_name}.', code=field_name)
        return value

    def get_lookup_value(self, model_obj=None, field_name=None):
        """Returns the field value by following the lookup string
        to a related instance.
        """
        value = model_obj
        lookup_string = self.lookups.get(field_name)
        for attr in lookup_string.split(LOOKUP_SEP):
            try:
                value = getattr(value, attr)
            except AttributeError:
                raise ValueGetterInvalidLookup(
                    f'Invalid lookup string. Got {lookup_string}')
        return value

    @property
    def m2m_field_names(self):
        """Returns the list of m2m field names for this model.
        """
        return [m2m.name for m2m in self.model_cls._meta.many_to_many]

    def get_m2m_value(self, model_obj=None, field_name=None):
        """Returns an m2m field value as a delimited string.
        """
        return self.m2m_delimiter.join(
            [value.name for value in getattr(model_obj, field_name).all()])

    def strip_value(self, value):
        """Returns a string cleaned of \n\t\r and double spaces.
        """
        try:
            value = value.replace(string.whitespace, ' ')
        except (TypeError, AttributeError):
            pass
        else:
            value = ' '.join(value.split())
        return value



from ..database import Database
from ..df_handlers import DfHandler
from .csv_exporter import CsvExporter


class CsvTablesExporterError(Exception):
    pass


class InvalidTableName(Exception):
    pass


class CsvTablesExporter:

    """Export to CSV all tables for an app_label.

    Usage:
        exporter = CsvTablesExporter(app_label='td')
        exporter.to_csv()
        exporter = CsvTablesExporter(app_label='edc')
        exporter.to_csv()

    """

    app_label = None
    export_folder = None
    csv_exporter_cls = CsvExporter
    db_cls = Database
    delimiter = '|'
    df_handler_cls = DfHandler
    exclude_history_tables = False
    excluded_app_labels = ['django_collect_offline']

    def __init__(self, app_label=None, with_columns=None, without_columns=None,
                 exclude_history_tables=None, exclude_table_hints=None,
                 export_folder=None, **kwargs):
        self.app_label = app_label or self.app_label
        if not self.app_label:
            raise CsvTablesExporterError(
                f'Missing app_label. Got None. See {repr(self)}')
        self.export_folder = export_folder or self.export_folder
        self.with_columns = with_columns or []
        self.without_columns = without_columns or []
        exclude_table_hints = exclude_table_hints or []
        self.exclude_history_tables = (
            self.exclude_history_tables
            if exclude_history_tables is None else exclude_history_tables)
        self.exported_paths = {}
        self.db = self.db_cls(**kwargs)
        self.table_names = self.get_table_names()
        for hint in (exclude_table_hints):
            for table_name in self.get_table_names():
                if hint in table_name:
                    self.table_names.pop(self.table_names.index(table_name))
        if self.exclude_history_tables:
            self.table_names = [
                tbl for tbl in self.table_names
                if 'historical' not in tbl and not tbl.endswith('_audit')]

    def __repr__(self):
        return f'{self.__class__.__name__}(app_label=\'{self.app_label}\')'

    def to_csv(self, table_names=None, export_folder=None, **kwargs):
        """Exports all tables to CSV.
        """
        self.exported_paths = {}
        export_folder = export_folder or self.export_folder
        if table_names:
            for table_name in table_names:
                if table_name not in self.table_names:
                    raise InvalidTableName(
                        f'{table_name} is not a valid for {self}')
            self.table_names = table_names
        for table_name in self.table_names:
            df = self.to_df(table_name=table_name, **kwargs)
            exporter = self.csv_exporter_cls(
                data_label=table_name,
                export_folder=export_folder,
                **kwargs)
            path = exporter.to_csv(
                dataframe=df, export_folder=export_folder)
            if path:
                self.exported_paths.update({table_name: path})

    def get_table_names(self):
        """Returns a list of table names for this app_label.
        """
        if self.with_columns:
            df = self.db.show_tables_with_columns(
                self.app_label, self.with_columns)
        elif self.without_columns:
            df = self.db.show_tables_without_columns(
                self.app_label, self.without_columns)
        else:
            df = self.db.show_tables(self.app_label)
        return list(df.table_name)

    def to_df(self, table_name=None, **kwargs):
        """Returns a dataframe after passing the raw df
        through the df_handler class.
        """
        df = self.get_raw_df(table_name)
        if self.df_handler_cls:
            df_handler = self.df_handler_cls(
                dataframe=df, db=self.db,
                table_name=table_name, **kwargs)
            df = df_handler.dataframe
        return df

    def get_raw_df(self, table_name=None):
        """Returns a df for the given table_name
        from an SQL statement, that is; raw).
        """
        return self.db.select_table(table_name=table_name)



import hashlib
import numpy as np
import pandas as pd

from .decrypt import decrypt


def identity(row, column_name=None):
    return decrypt(row, column_name or 'identity', 'rsa', 'restricted')


def identity256_decrypt(row, column_name=None):
    identity256 = np.nan
    column_name = column_name or 'identity'
    if pd.notnull(row[column_name]):
        identity256 = hashlib.sha256(
            identity(row, column_name)).digest().encode("hex")
    return identity256


def identity256(row, column_name=None):
    identity256 = np.nan
    column_name = column_name or 'identity'
    if pd.notnull(row[column_name]):
        identity256 = hashlib.sha256(
            row[column_name].encode('utf-8')).hexdigest()
    return identity256





class RsDialect:

    def __init__(self, obj=None):
        self.obj = obj

    @property
    def select_registered_subject(self):
        """Returns an SQL statement that joins visit, appt, and registered_subject.
        """
        sql = (
            'SELECT R.subject_identifier, R.screening_identifier, R.dob, '
            'R.gender, R.subject_type, R.sid, '
            'R.screening_age_in_years, R.registration_status, R.registration_datetime, '
            'R.randomization_datetime '
            f'FROM {self.obj.registered_subject_tbl} as R '
        )
        return sql, None



from django.contrib import admin

from ..admin_site import edc_pdutils_admin
from ..models import DataRequest


@admin.register(DataRequest, site=edc_pdutils_admin)
class DataRequestAdmin(admin.ModelAdmin):

    fields = ('requested', 'export_format', 'decrypt', )

    list_display = ('archive_filename', 'user_created', 'created')



from django.db import models
from django.db.models.deletion import PROTECT
from django_crypto_fields.fields.encrypted_char_field import EncryptedCharField
from edc_appointment.models import Appointment
from edc_base.model_mixins import BaseUuidModel
from edc_base.model_mixins.list_model_mixin import ListModelMixin
from edc_base.utils import get_utcnow
from edc_constants.constants import YES


class SubjectVisit(BaseUuidModel):

    appointment = models.OneToOneField(Appointment, on_delete=models.PROTECT)

    subject_identifier = models.CharField(max_length=25)

    report_datetime = models.DateTimeField(default=get_utcnow)

    visit_code = models.CharField(max_length=25, default='T0')

    reason = models.CharField(max_length=25, null=True)

    survival_status = models.CharField(max_length=25, null=True)

    last_alive_date = models.DateTimeField(null=True)

    class Meta:
        ordering = ['report_datetime']


class SubjectConsent(BaseUuidModel):

    subject_identifier = models.CharField(max_length=25)

    consent_datetime = models.DateTimeField(default=get_utcnow)

    dob = models.DateField(null=True)

    citizen = models.CharField(max_length=25, default=YES)

    legal_marriage = models.CharField(max_length=25, null=True)

    marriage_certificate = models.CharField(max_length=25, null=True)

    happy = models.CharField(
        max_length=25,
        choices=(
            ('YES', 'YES'), ('NO', 'NO')),
        null=True)


class SubjectLocator(BaseUuidModel):

    subject_identifier = models.CharField(max_length=25)


class CrfModelMixin(models.Model):

    subject_visit = models.OneToOneField(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(null=True)

    @property
    def subject_identifier(self):
        return self.subject_visit.subject_identifier

    @property
    def visit_code(self):
        return self.subject_visit.visit_code

    @property
    def visit(self):
        return self.subject_visit

    class Meta:
        abstract = True


class SubjectRequisition(CrfModelMixin, BaseUuidModel):

    panel_name = models.CharField(max_length=25, default='Microtube')


class ListModel(ListModelMixin):
    pass


class Crf(CrfModelMixin, BaseUuidModel):

    char1 = models.CharField(max_length=25, null=True)

    date1 = models.DateTimeField(null=True)

    int1 = models.IntegerField(null=True)

    uuid1 = models.UUIDField(null=True)

    m2m = models.ManyToManyField(ListModel)


class CrfOne(CrfModelMixin, BaseUuidModel):

    dte = models.DateTimeField(default=get_utcnow)


class CrfTwo(CrfModelMixin, BaseUuidModel):

    dte = models.DateTimeField(default=get_utcnow)


class CrfThree(CrfModelMixin, BaseUuidModel):

    UPPERCASE = models.DateTimeField(default=get_utcnow)


class CrfInline(BaseUuidModel):

    crf_one = models.ForeignKey(CrfOne, on_delete=models.PROTECT)

    crf_two = models.ForeignKey(CrfTwo, on_delete=models.PROTECT)

    dte = models.DateTimeField(default=get_utcnow)


class CrfEncrypted(CrfModelMixin, BaseUuidModel):

    encrypted1 = EncryptedCharField(null=True)
